package io.irandom.messaging; 

import java.net.*;
import java.io.*;
import java.util.StringTokenizer;
import java.util.Hashtable;
import java.util.Vector;
import salsa.language.*;
import io.core.messaging.*;
import io.core.profiling.DecisionAgent;
import io.core.server.*;
import io.random.messaging.RandomBasicHandler;
import io.random.messaging.RandomProtocolHandler;
import io.irandom.profiling.*;

public class IRandomPeerTheater implements IOTheater {
	// Must be a singleton so UniversalActor can get to it.
	private static IRandomPeerTheater instance = new IRandomPeerTheater();

	private static Hashtable messageQueue;
	private static Hashtable registry;
	private static Vector neighbors;
	private static Vector peerServerList;

	private static ServerSocket passive;
	private static int listenPort = 3030;

	private static TheaterLocator location;

	private static IRandomDecisionAgent decider = new IRandomDecisionAgent(instance);
	private static RandomProtocolHandler protocolHandler = new RandomProtocolHandler(instance);
	private static RandomBasicHandler basicHandler = new RandomBasicHandler(instance);

	private static int maxConnections = 10;
	private static double processing = 0;
	private static double bandwidth = 0;

	/**
		This method creates a passive socket and starts accepting connections to it.
	*/
	private void listen() {
		try {
			passive = new ServerSocket( listenPort );
		} catch (IOException e) {
        		System.err.println( "IOException creating server socket. " + e.getMessage() );

			return;
		}

		listenPort = passive.getLocalPort();
		System.out.println( "Theater listening at: " + location.toString() );

		// Loop and accept incoming connections.
		while (true) {
			Socket incoming = null;

		         // Accept the connection.
			try {
				incoming = passive.accept();
			} catch (IOException e) {
				System.err.println( "IOException occured accepting new connection" + e.getMessage() );

				return;
			}

			//Handle the incoming connection
			ObjectHandler handler = new ObjectHandler(incoming, this);
			handler.start();
		}
	}

	public static void main(String[] args) {
		// Handle arguments...
		getSettings(args);

		if ((processing <= 0) && (bandwidth <= 0)) {
			System.err.println("IRandomPeerServer requires positive values for bandwidth and processing");
			System.exit(0);
		}


		try {
			InetAddress IPHere = InetAddress.getLocalHost();

			String StringHere = "rmsp://" + IPHere.getHostAddress() + ":" + listenPort;
			location = new TheaterLocator(StringHere);
		} catch (Exception e) {
			System.err.println("Error resolving localhost: " + e);
			System.exit(0);
		}

		decider.start();

		//	The PeerConnector will connect to the given servers and look for 
		//	The Decider will decide when to migrate actors to other peer theaters, and when to
		//	accept actors from other peer theaters
		if (peerServerList.size() > 0) {
			ServerHandler serverHandler = new ServerHandler(instance, peerServerList, maxConnections);
			serverHandler.start();
		}

         	instance.listen();
   	}

	private static void getSettings(String[] args) {
		for (int i = 0; i < args.length; i++) {

			if (args[i].equals("-?") || args[i].equals("-h")) 	IRandomPeerTheater.helpMessage();
			else if (args[i].equals("-p") || args[i].equals("-c") || args[i].equals("-f") || args[i].equals("-pr") || args[i].equals("-bw")) {
				try {
					if (args[i].equals("-p")) 		listenPort = Integer.parseInt(args[i+1]);
					else if (args[i].equals("-pr"))		processing = Double.parseDouble(args[i+1]);
					else if (args[i].equals("-bw"))		bandwidth = Double.parseDouble(args[i+1]);
					else if (args[i].equals("-c"))		maxConnections = Integer.parseInt(args[i+1]);
					else if (args[i].equals("-f"))		getSettingsFromFile(args[i+1]);
				} catch (Exception e) {
					if (args[i].equals("-p")) 		IRandomPeerTheater.helpMessage("invalid port number : " + args[i+1]);
					else if (args[i].equals("-pr"))		IRandomPeerTheater.helpMessage("invalid processing value : " + args[i+1]);
					else if (args[i].equals("-bw"))		IRandomPeerTheater.helpMessage("invalid bandwidth value : " + args[i+1]);
					else if (args[i].equals("-c"))		IRandomPeerTheater.helpMessage("invalid number of max connections : " + args[i+1]);
				}
				i = i + 1;
			}
			else if (args[i].equals("-s")) {
				i++;
				peerServerList = new Vector();
				while ((i < args.length) && ((args[i]).charAt(0) != '-')) {
					peerServerList.add(args[i]);							
					i++;
				}
				i--;
				if (args[i].equals("-s")) IRandomPeerTheater.helpMessage("You must specify at least one PeerServer");

			}
			else IRandomPeerTheater.helpMessage("Invalid Arguments");
		}
	}


	private static void getSettingsFromFile(String filename) {
		FileInputStream file;
		byte input[];

		try {
			file = new FileInputStream(filename);

			input = new byte[file.available()];
			file.read(input);
		} catch (Exception e) {
			System.err.println("File Error: " + e);
			return;
		}

		StringTokenizer st = new StringTokenizer(new String(input), "\n ");
		String args[] = new String[st.countTokens()];

		int i = 0;
		while (st.hasMoreTokens()) {
			args[i] = st.nextToken();
			i++;
		}

		getSettings(args);
	}


	private static void helpMessage(String s) {
		System.out.println(s);
		IRandomPeerTheater.printHelpMessage();
		System.exit(1);
	}

	private static void helpMessage() {
		IRandomPeerTheater.printHelpMessage();
		System.exit(0);
	}

	private static void printHelpMessage() {
		System.out.println("usage :");
		System.out.println("\tjava ...IRandomPeerTheater");
		System.out.println("\tjava ...IRandomPeerTheater -h");
		System.out.println("\tjava ...IRandomPeerTheater -v");
		System.out.println("\tjava ...IRandomPeerTheater -p   portNumber");
		System.out.println("\t                           -s   server1 server2 ... serverN");
		System.out.println("\t				 -pr  processing");
		System.out.println("\t				 -bw  bandwidth");
		System.out.println("\t                           -c   numConnections");
		System.out.println("\t                           -f   filename");

		System.out.println("options :");
		System.out.println("\t-h                             : print this message");
		System.out.println("\t-v                             : print version");
		System.out.println("\t-pr			     : available processing for peer theater");
		System.out.println("\t-bw			     : available bandwidth for peer theater");
		System.out.println("\t-f filename                    : load settings from file");
		System.out.println("\t                                 file is in the same format as command line arguments");
		System.out.println("\t-p portNumber                  : set the listening port to portNumber");
		System.out.println("\t                                 default port number is 3030");
		System.out.println("\t-s server1 server2 ... serverN : gets connections to other");
		System.out.println("\t                                 peer theaters, from servers 1 through N");
		System.out.println("\t-c numConnections              : specifies the maximum amount of connections");
		System.out.println("\t                                 kept to peer theaters, default is 10");
	}

	/**
		The following functions are used by autonomous and universal actors to access
		various information found in the theater
	*/
	public static IRandomPeerTheater getInstance() {
		return instance;
	}

	public DecisionAgent getDecisionAgent() {
		return decider;
	}

	public ProtocolHandler getProtocolHandler() {
		return protocolHandler;
	}

	public BasicHandler getBasicHandler() {
		return basicHandler;
	}

	public TheaterLocator getLocation() {
		return location;
	}

	public Hashtable getRegistry() {
		return registry;
	}

	public Hashtable getMessageQueue() {
		return messageQueue;
	}

	public Vector getNeighbors() {
		return neighbors;
	}

	public double getProcessing() {
		return processing;
	}

	public double getBandwidth() {
		return bandwidth;
	}

	private IRandomPeerTheater() {
		registry = new Hashtable();
		messageQueue = new Hashtable();
		neighbors = new Vector();
		peerServerList = new Vector();
	}

   	public int getPort() {
		return listenPort;
	}

   	public UniversalActor getActorUAL( String relativeUAL ) {
		//This returns a UniversalActor, given a particular UAL
		return (UniversalActor)registry.get(relativeUAL);
	}
}
